/*
 * squareExplore.c
 *
 * Robert Hochberg
 * April 10, 2012
 *
 * Find rational periodic points of the x -> x^2+a iteration
 */

#include "seMouseRotation.h"


/* 
 * This function returns a double approximating how large the 
 * denominator of the input number, supposed to be less than 1,
 * would be if the number were rational. In some sense...
 */
__device__ double denomVal(double a, double b, int depth){
  unsigned long long int intPart, start1, start2;
  int d, numIterations;
  a = fabs(a);
  b = fabs(b);
  unsigned long long int rval1 = 0, rval2 = 0;

  // Partial fractions analysis on a
  numIterations = 0;
  intPart = (unsigned long long int)a;
  start1 = 0;
  start2 = 1;
  d = 0;
  do{
    d = intPart*start2 + start1;
    start1 = start2;
    start2 = d;
    numIterations++;
    a = 1.0/(a-intPart);
    intPart = (unsigned long long int)a;
    rval1 = max(rval1, intPart);
  }  while(intPart < (1<<depth) && numIterations < 5*depth);

  // Repeat for b
  numIterations = 0;
  intPart = (unsigned long long int)b;
  start1 = 0;
  start2 = 1;
  d = 0;
  //  while(b != intPart && numIterations <= depth){
  do{
    d = intPart*start2 + start1;
    start1 = start2;
    start2 = d;
    numIterations++;
    b = 1.0/(b-intPart);
    intPart = (unsigned long long int)b;
    rval2 = max(rval2, intPart);
  }while(intPart < (1<<depth) && numIterations < 5*depth);

  return logf(rval1*rval1 + rval2*rval2);
}


/*
 * This kernel function fills a device array dResults[] with the
 * values of f(f(...f(x)...)) - x, where the function f(x) = x^2+a,
 * and the iteration is performed numIt times.
 *
 * The starting values x for the iteration are xmin + i*xstep, up to xmax,
 * The values for a are amin + j*astep, up to amax.
 */
__global__ void iterate(double xmin, double xmax, double xstep, 
			double amin, double amax, double astep, 
			int numIt, size_t stride, double* dResults,
			int depth, double* dDenomVals){
  int xidx = blockIdx.x * blockDim.x + threadIdx.x;
  int aidx = blockIdx.y * blockDim.y + threadIdx.y;
  double xval = xmin + xidx * xstep;
  double aval = amin + aidx * astep;
  double originalXval = xval;

  // Make sure we're within range
  if(xval > xmax || aval > amax)
    return;

  // Iterate
  for(int i = 0; i < numIt; i++)
    xval = xval * xval + aval; 

  // Save our value to global memory
  int loc = aidx * stride + xidx;

  dResults[loc] = xval - originalXval;
  dDenomVals[loc] = denomVal(fabs(xval - originalXval), fabs(aval), depth);
}



void drawString (void * font, char *s, float x, float y, float z){
     unsigned int i;
     glRasterPos3f(x, y, z);
     for (i = 0; i < strlen (s); i++)
          glutBitmapCharacter (font, s[i]);
} 

void lightDisplay(void){
  GLfloat mat_spec[] = {0.0, 0.0, 0.1, 1.0};
  GLfloat mat_shininess[] = {100.0};
  GLfloat light_position[] = {0.5, 0.5, 0.5, 4.0};
  GLfloat light_specular[] = {1.0, 1.0, 1.0, 1.0};
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_spec);
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
}

/* This generates text for the subwindow textWindow.  
 *
 * It is also the displayForGlut method but for textWindow.
 * Meaning textWindow callback for display is displayText
 * not displayForGlut.
 */
void displayText(void){
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glColor3f(1.0, 1.0, 1.0);
  sprintf(gIterationLoc, "Iteration: %d   Depth: %0.2f", iterations, gdepth);
  sprintf(gMouseXLoc, "x: %2.20f", xMotionLoc);
  sprintf(gMouseALoc, "a: %2.20f", yMotionLoc); 
  sprintf(gFracX, "x Frac: %lld/%lld", fracEntry(xMotionLoc, NUMERATOR, gdepth),
         fracEntry(xMotionLoc, DENOMINATOR, gdepth)); 
  sprintf(gFracA, "a Frac: %lld/%lld", fracEntry(yMotionLoc, NUMERATOR, gdepth),
         fracEntry(yMotionLoc, DENOMINATOR, gdepth)); 
  drawString(GLUT_BITMAP_HELVETICA_18, gMouseXLoc, 0, 0.65, 0);
  drawString(GLUT_BITMAP_HELVETICA_18, gMouseALoc, 0, 0.45, 0);
  drawString(GLUT_BITMAP_HELVETICA_18, gIterationLoc, 0, 0.85, 0);
  drawString(GLUT_BITMAP_HELVETICA_18, gFracX, 0, 0.25, 0);
  drawString(GLUT_BITMAP_HELVETICA_18, gFracA, 0, 0.05, 0);
  glutPostRedisplay();
  glutSwapBuffers();
}

/*
 * This generates a picture of the values in the hResults array.
 *
 * It is assumed that hResults is filled with meaningful data
 * prior to this functi on call.
 */
void displayForGlut(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  lightDisplay();
  double pw = 1.0/xcount;
  double ph = 1.0/acount;
  for(int i = 0; i < acount-1; i++){
    double y = ((double)i) / acount;
    for(int j = 0; j < xcount-1; j++){
      double x = ((double)j) / xcount;
      if(fabs(hResults[i * stride + j]) < threshold &&
	 fabs(hResults[i * stride + j + 1]) < threshold &&
	 fabs(hResults[(i + 1) * stride + j]) < threshold &&
	 fabs(hResults[(i + 1) * stride + j + 1]) < threshold){
	double r = (hDenomVals[i * stride + j]);
	glColor3f(1/r, fabs(sin(r+1.79*gcolor)), fabs(cos(r+gcolor)));

        glBegin(GL_QUADS);
	glVertex3f(x, y, hResults[i * stride + j]/threshold/7);
        glVertex3f(x+pw, y, hResults[i * stride + j + 1]/threshold/7);
        glVertex3f(x+pw, (y+ph), hResults[(i + 1) * stride + j+1]/threshold/7);
        glVertex3f(x, (y+ph), hResults[(i + 1) * stride + j]/threshold/7);

        glEnd();
      }
    }
  }
  glutSwapBuffers();
}

/* Handles all non-special keyboard functions
 *
 */
void kbdForGlut(unsigned char key, int x, int y){
  double gxCenter = (gxmin + gxmax)/2;
  double gaCenter = (gamin + gamax)/2;

  switch(key){
  case (unsigned char)3:
  case 'q':
    exit(0);
    break;
  case 'f':
    glTranslatef(0.0, 0.5, 0.0);
    glRotatef(5.0, -1.0, 0.0, 0.0);
    glTranslatef(0.0, -0.5, 0.0);
    break;
  case 'F':
    glTranslatef(0.0, 0.5, 0.0);
    glRotatef(-5.0, -1.0, 0.0, 0.0);
    glTranslatef(0.0, -0.5, 0.0);
    break;
  case 'g':
    glTranslatef(0.5, 0.0, 0.0);
    glRotatef(5.0, 0.0, 1.0, 0.0);
    glTranslatef(-0.5, 0.0, 0.0);
    break;
  case 'G':
    glTranslatef(0.5, 0.0, 0.0);
    glRotatef(-5.0, 0.0, 1.0, 0.0);
    glTranslatef(-0.5, 0.0, 0.0);
    break;
  case 't':
    threshold = threshold * 2;
    break;
  case 'T':
    threshold = threshold / 2;
    break;
  case 'd':
    gdepth = gdepth + 2;
    break;
  case 'D':
    gdepth = gdepth - 2;
    if(gdepth < 0)
      gdepth = 0;
    break;
  case 'c':
    gcolor = gcolor + 0.37882847;
    break;
  case 'C':
    gcolor = gcolor - 0.37882847;
    break;
  case 'i':
    iterations = iterations + 1;
    break;
  case 'I':
    if(iterations >= 1)
      iterations = iterations - 1;
    break;
  case 'z':
    gxmin = (gxmin + gxCenter) / 2;
    gxmax = (gxmax + gxCenter) / 2;
    gamin = (gamin + gaCenter) / 2;
    gamax = (gamax + gaCenter) / 2;
    gastep /= 2;
    gxstep /= 2;
    break;
  case 'Z':
    gxmin = 2*gxmin - gxCenter;
    gxmax = 2*gxmax - gxCenter;
    gamin = 2*gamin - gaCenter;
    gamax = 2*gamax - gaCenter;
    gastep *= 2;
    gxstep *= 2;
    break;
  case 'r':			/* reset */
    glLoadIdentity();
    gxmin = -5.0;
    gxmax = 5.0;
    gxstep = 0.05;
    gamin = -5.0;
    gamax = 5.0;
    gastep = 0.05;
    threshold = 0.1;
    gdepth = 3;
    break;
  case 's':
    gxstep = gxstep / 2;
    if((gxmax-gxmin)/gxstep > 1024) gxstep = (gxmax - gxmin) / 1024.0;
    gastep = gastep / 2;
    if((gamax-gamin)/gastep > 1024) gastep = (gamax - gamin) / 1024.0;
    break;
  case 'S':
    gxstep = gxstep * 2;
    if((gxmax-gxmin)/gxstep < 4) gxstep = (gxmax - gxmin) / 4.0;
    gastep = gastep * 2;
    if((gamax-gamin)/gastep < 4) gastep = (gamax - gamin) / 4.0;
    break;
  case 'o':
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glFrustum(0.0, 1.0, 0.0, 1.0, -10.0, 10.0);
    break;
  }
  
  
  recompute(gxmin, gxmax, gxstep, gamin, gamax, gastep, iterations, stride, (double*)dResults);
  glutPostRedisplay();
}

void mouseMotionForGlut(int x, int y){
  double xloc = gxmin + (gxmax-gxmin)*((double)x)/glutGet(GLUT_WINDOW_WIDTH);
  double yloc = gamin + (gamax-gamin)*(1.0-((double)y)/glutGet(GLUT_WINDOW_HEIGHT));
  if(g_bButton1Down){
    printf("xloc: %lf, yloc: %lf\n", xloc, yloc);
    if(xloc < xMotionLoc){
      glTranslatef(0.5, 0.0, 0.0);
      glRotatef(2, 0.0, 1.0, 0.0);
      glTranslatef(-0.5, 0.0, 0.0);
    }
    if(xloc > xMotionLoc){
      glTranslatef(0.5, 0.0, 0.0);
      glRotatef(-2, 0.0, 1.0, 0.0);
      glTranslatef(-0.5, 0.0, 0.0);
    }
    if(yloc < yMotionLoc){
      glTranslatef(0.0, 0.5, 0.0);
      glRotatef(-2, -1.0, 0.0, 0.0);
      glTranslatef(0.0, -0.5, 0.0);
    }
    if(yloc > yMotionLoc){
      glTranslatef(0.0, 0.5, 0.0);
      glRotatef(2, -1.0, 0.0, 0.0);
      glTranslatef(0.0, -0.5, 0.0);
    }
  }

  xMotionLoc = gxmin + (gxmax-gxmin)*((double)x)/glutGet(GLUT_WINDOW_WIDTH);
  yMotionLoc = gamin + (gamax-gamin)*(1.0-((double)y)/glutGet(GLUT_WINDOW_HEIGHT));
  glutPostRedisplay();
}

void mouseForGlut(int button, int state, int x, int y){
  if (button == GLUT_LEFT_BUTTON){
    g_bButton1Down = (state == GLUT_DOWN) ? true : false;
    if(g_bButton1Down){
      xClickedLocation = gxmin + (gxmax-gxmin)*((double)x)/glutGet(GLUT_WINDOW_WIDTH);
      yClickedLocation = gamin + (gamax-gamin)*(1.0-((double)y)/glutGet(GLUT_WINDOW_HEIGHT));
    }
  }
}

/*
 * Uses continued fractions to find numerators and denominators
 * of small fractions that aproximate the input double a.
 * 
 * "fracpos" is an enumerated type {NUMERATOR, DENOMINATOR}
 * telling the function to return either the numerator or the
 * denominator of the fraction.
 */
unsigned long long int fracEntry(double a, fracpos tob, int depth){
  unsigned long long int numIterations, intPart, start1, start2, d;
  int sign = 1;
  // Partial fractions analysis on a
  if(a < 0){
    a = fabs(a);
    if(tob == NUMERATOR)
      sign = -1;
  }

  numIterations = 0;
  intPart = (unsigned long long int)a;
  if(tob == NUMERATOR){
    start1 = 0; 
    start2 = 1;
  } else {
    start1 = 1;
    start2 = 0;
  }

  do{
    d = intPart*start2 + start1;
    start1 = start2;
    start2 = d;
    numIterations++;
    a = 1.0/(a-intPart);
    intPart = (unsigned long long int)a;
  } while(intPart < 50  && numIterations < 5*depth);
  return sign * start2;
}

void arrowsForGlut(int key, int x, int y){
  double factor = 4.0;
  double xoffset = (gxmax - gxmin)/factor;
  double aoffset = (gamax - gamin)/factor;
  switch(key){
  case GLUT_KEY_LEFT:
    gxmax = gxmax - xoffset;
    gxmin = gxmin - xoffset;
    break;
  case GLUT_KEY_RIGHT:
    gxmax = gxmax + xoffset;
    gxmin = gxmin + xoffset;
    break;
  case GLUT_KEY_DOWN:
    gamax = gamax - aoffset;
    gamin = gamin - aoffset;
    break;
  case GLUT_KEY_UP:
    gamax = gamax + aoffset;
    gamin = gamin + aoffset;
    break;
  default:
    break;
  }
  recompute(gxmin, gxmax, gxstep, gamin, gamax, gastep, iterations, stride, (double*)dResults);
  glutSwapBuffers();
  glutPostRedisplay();
}


int initGlutDisplay(int argc, char* argv[]){
  glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);
  glutInitWindowSize (800, 800);
  glutInitWindowPosition (100, 100);
  mainWindow = glutCreateWindow ("SquareExplorer - CUDA");
  glClearColor (0.0, 0.0, 0.0, 0.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, 1.0, 0.0, 1.0, -10.0, 10.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // setting Depth
  glClearDepth(1.0f);
  glEnable(GL_DEPTH_TEST);
  // setting light 
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_COLOR_MATERIAL);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  //Callback registration
  glutDisplayFunc(displayForGlut);
  glutKeyboardFunc(kbdForGlut);
  glutSpecialFunc(arrowsForGlut);
  glutMotionFunc(mouseMotionForGlut);
  glutPassiveMotionFunc(mouseMotionForGlut);
  glutMouseFunc(mouseForGlut);

  //creating subwindow for text 
  textWindow = glutCreateSubWindow(mainWindow, 1, 1, 250, 100);
  glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0);
  glClearColor (0.0, 0.0, 0.0, 0.01);
  glutDisplayFunc(displayText);
  glutPassiveMotionFunc(mouseMotionForGlut);
  glutKeyboardFunc(kbdForGlut);
  glutMainLoop();
  return 0;
}


/* 
 * This function fills the array with data, based on the values
 * of the parameters, which define the window and step.
 */
 void recompute(double xmin, double xmax, double xstep,
		double amin, double amax, double astep,
		int numIt, size_t stride, double* dResults){

  // How many entries do we need?
  xcount = (size_t)floor( (xmax - xmin)/xstep );
  acount = (size_t)floor( (amax - amin)/astep );

  // Compute block and grid size
  dim3 dimBlock(BLOCKSIZE, BLOCKSIZE);
  dim3 dimGrid((int)ceil((xmax - xmin)/xstep), (int)ceil((amax - amin)/astep));

  iterate <<< dimGrid, dimBlock >>> (xmin, xmax, xstep, amin, amax, astep, 
				     numIt, stride, (double*)dResults, gdepth, (double*)dDenomVals);
  cudaError_t err = cudaThreadSynchronize();
  printf("After recompute Call: %s\n", cudaGetErrorString(err));

  // Copy the results off of the device onto the host
  err = cudaMemcpy(hResults, dResults, stride * acount * sizeof(double), cudaMemcpyDeviceToHost);
  printf("Copied results to host: %s\n", cudaGetErrorString(err));
  err = cudaMemcpy(hDenomVals, dDenomVals, stride * acount * sizeof(double), cudaMemcpyDeviceToHost);
  printf("Copied results to host: %s\n", cudaGetErrorString(err));

 }



/*
 * We explore a rectangle of a-values and x-values, where the x-values
 * are the starting points, and the a values are the constants in the 
 * iteration x -> x^2 + a.
 */
int main(int argc, char* argv[]){
  gxmin = -5.0;
  gxmax = 5.0;
  gxstep = 0.05;
  gamin = -5.0;
  gamax = 5.0;
  gastep = 0.05;
  threshold = 0.1;
  iterations = 1;
  gdepth = 3;			/* depth of continued fraction iteration */
  g_bButton1Down = false;

  // How many entries do we need?
  xcount = (size_t)floor( (gxmax - gxmin)/gxstep );
  acount = (size_t)floor( (gamax - gamin)/gastep );

  // Let's make room for the most we'll ever have.
  // Limit to 1024 values in each direction, per computation
  // Allocate memory on device
  cudaError_t err =  cudaMallocPitch((void**)&dResults, &stride, 
				     1088*sizeof(double), 1088*sizeof(double));
  err =  cudaMallocPitch((void**)&dDenomVals, &stride, 
				     1088*sizeof(double), 1088*sizeof(double));
 
  stride = stride / sizeof(double);

  printf("Allocate: x:%d, a:%d, Stride = %d: %s\n", (int)xcount, (int)acount, 
	 (int)stride, cudaGetErrorString(err));

  // We reserve enough room for our full-size array: 1024x1024 doubles = 8MB
  // Allocate memory on the host (Note we reserve a bit more...)
  hResults = (double*)malloc((size_t)(stride * 1088 * sizeof(double)));
  hDenomVals = (double*)malloc((size_t)(stride * 1088 * sizeof(double)));

  // Compute block and grid size
  dim3 dimBlock(BLOCKSIZE, BLOCKSIZE);
  dim3 dimGrid((int)ceil((gxmax - gxmin)/gxstep), (int)ceil((gamax - gamin)/gastep));

  iterate <<< dimGrid, dimBlock >>> (gxmin, gxmax, gxstep, gamin, gamax, gastep, 
				     iterations, stride, (double*)dResults, 
				     gdepth, (double*)dDenomVals);
  err = cudaThreadSynchronize();
  printf("After Kernel Call: %s\n", cudaGetErrorString(err));

  // Copy the results off of the device onto the host
  err = cudaMemcpy(hResults, dResults, stride * acount * sizeof(double), cudaMemcpyDeviceToHost);
  printf("Copied results to host: %s\n", cudaGetErrorString(err));
  err = cudaMemcpy(hDenomVals, dDenomVals, stride * acount * sizeof(double), cudaMemcpyDeviceToHost);
  printf("Copied denomVals to host: %s\n", cudaGetErrorString(err));



  // Get the graphics up and running
  initGlutDisplay(argc, argv);

  // Now print out the array
  for(int i = 0; i < acount; i++){
    for(int j = 0; j < xcount; j++){
      printf("%f\t", hResults[i * stride + j]);
    }
    printf("\n");
  }
  printf("\n");
}
